
from . access import ProfileAccessModel

from anthill.common import access, profile
from anthill.common.profile import ProfileError, FuncError, NoDataError
from anthill.common.model import Model
from anthill.common.database import DatabaseError, format_conditions_json, ConditionError

import ujson


class NoSuchProfileError(Exception):
    pass


class ProfileQueryError(Exception):
    pass


class ProfileAdapter(object):
    def __init__(self, data):
        self.account = str(data.get("account_id"))
        self.profile = data.get("payload")


class ProfileQuery(object):
    def __init__(self, gamespace_id, db):
        self.gamespace_id = gamespace_id
        self.db = db

        self.filters = None

        self.offset = 0
        self.limit = 0

    def __values__(self):
        conditions = [
            "`account_profiles`.`gamespace_id`=%s"
        ]

        data = [
            str(self.gamespace_id)
        ]

        if self.filters:
            for condition, values in format_conditions_json('payload', self.filters):
                conditions.append(condition)
                data.extend(values)

        return conditions, data

    async def query(self, one=False, count=False):
        try:
            conditions, data = self.__values__()
        except ConditionError as e:
            raise ProfileQueryError("Failed to process profile conditions: {0}".format(str(e)))

        query = """
            SELECT {0} `account_id`, `payload` FROM `account_profiles`
            WHERE {1}
        """.format(
            "SQL_CALC_FOUND_ROWS" if count else "",
            " AND ".join(conditions))

        if self.limit:
            query += """
                LIMIT %s,%s
            """

            data.append(int(self.offset))
            data.append(int(self.limit))

        query += ";"

        if one:
            try:
                result = await self.db.get(query, *data)
            except DatabaseError as e:
                raise ProfileQueryError("Failed to get profiles: " + e.args[1])

            if not result:
                return None

            return ProfileAdapter(result)
        else:
            try:
                result = await self.db.query(query, *data)
            except DatabaseError as e:
                raise ProfileQueryError("Failed to query profiles: " + e.args[1])

            count_result = 0

            if count:
                count_result = await self.db.get(
                    """
                        SELECT FOUND_ROWS() AS count;
                    """)
                count_result = count_result["count"]

            items = map(ProfileAdapter, result)

            if count:
                return items, count_result

            return items


class ProfilesModel(Model):
    TIME_CREATED = "@time_created"
    TIME_UPDATED = "@time_updated"

    # noinspection PyShadowingNames
    def __init__(self, db, access):
        self.db = db
        self.access = access

    def get_setup_tables(self):
        return ["account_profiles"]

    def get_setup_db(self):
        return self.db

    def has_delete_account_event(self):
        return True

    async def accounts_deleted(self, gamespace, accounts, gamespace_only):
        if gamespace_only:
            await self.db.execute(
                """
                    DELETE FROM `account_profiles`
                    WHERE `gamespace_id`=%s AND `account_id` IN %s;
                """, gamespace, accounts)
        else:
            await self.db.execute(
                """
                    DELETE FROM `account_profiles`
                    WHERE `account_id` IN %s;
                """, accounts)

    async def delete_profile(self, gamespace_id, account_id):
        await self.db.execute(
            """
                DELETE FROM `account_profiles`
                WHERE `account_id`=%s AND `gamespace_id`=%s;
            """, account_id, gamespace_id)

    def profile_query(self, gamespace_id):
        return ProfileQuery(gamespace_id, self.db)

    async def get_profile_data(self, gamespace_id, account_id, path):
        user_profile = UserProfile(
            self.db,
            gamespace_id,
            account_id)

        try:
            data = await user_profile.get_data(path)
        except NoDataError:
            raise NoSuchProfileError()

        return data

    async def get_profile_me(self, gamespace_id, account_id, path):
        
        profile_data = await self.get_profile_data(gamespace_id, account_id, path)

        if not path:
            # if the path is not specified, get them all
            valid_keys = await self.access.validate_access(
                gamespace_id,
                list(profile_data.keys()),
                ProfileAccessModel.READ)

            result = {
                key: profile_data[key]
                for key in valid_keys
            }
        else:

            key = path[0]
            valid_keys = await self.access.validate_access(
                gamespace_id,
                [key],
                ProfileAccessModel.READ)

            if valid_keys:
                result = profile_data
            else:
                result = None

        return result

    async def get_profile_others(self, gamespace_id, account_id, path):
        
        profile_data = await self.get_profile_data(gamespace_id, account_id, path)

        if not path:
            # if the path is not specified, get them all
            valid_keys = await self.access.validate_access(
                gamespace_id,
                list(profile_data.keys()),
                ProfileAccessModel.READ_OTHERS)

            result = {
                key: profile_data[key]
                for key in valid_keys
            }
        else:
            key = path[0]
            valid_keys = await self.access.validate_access(
                gamespace_id,
                [key],
                ProfileAccessModel.READ_OTHERS)

            if valid_keys:
                result = profile_data
            else:
                result = None

        return result

    async def get_profiles(self, gamespace_id, action, account_ids, profile_fields):

        async def get_private():

            result = {}

            for account_id in account_ids:
                data = (await self.get_profile_data(gamespace_id, account_id, [])) or {}

                if profile_fields:
                    result[account_id] = {
                        field: (data[field])
                        for field in profile_fields if field in data
                    }
                else:
                    result[account_id] = data

            return result

        async def get_public():

            if profile_fields:
                valid_fields = await self.access.validate_access(gamespace_id, profile_fields,
                                                                 ProfileAccessModel.READ_OTHERS)
            else:
                public_access = await self.access.get_access(gamespace_id)
                valid_fields = public_access.get_public()

            result = {}

            for account_id in account_ids:
                try:
                    data = (await self.get_profile_data(gamespace_id, account_id, []))
                except NoSuchProfileError:
                    data = {}
                result[account_id] = {
                    field: (data[field])
                    for field in valid_fields if field in data
                }

            return result

        actions = {
            "get_private": get_private,
            "get_public": get_public
        }

        if action not in actions:
            raise ProfileError("No such profile action: " + action)
        if len(account_ids) > 1000:
            raise ProfileError("Maximum account limit exceeded (1000).")
        profiles = await actions[action]()
        return profiles

    async def set_profile_data(self, gamespace_id, account_id, fields, path, merge=True):
        user_profile = UserProfile(self.db, gamespace_id, account_id)
        try:
            result = await user_profile.set_data(fields, path, merge=merge)
        except FuncError as e:
            raise ProfileError("Failed to update profile: " + e.message)
        return result

    async def set_profiles_data(self, gamespace_id, accounts: dict, merge=True):
        user_profiles = UserProfiles(self.db, gamespace_id, list(accounts.keys()))
        try:
            result = await user_profiles.set_data(accounts, None, merge=merge)
        except FuncError as e:
            raise ProfileError("Failed to update profiles: " + e.message)
        return result

    async def set_profile_me(self, gamespace_id, account_id, fields, path, merge=True):

        if not path:
            await self.access.validate_access(
                gamespace_id,
                list(fields.keys()),
                ProfileAccessModel.WRITE)

            result = await self.set_profile_data(gamespace_id, account_id, fields, path, merge=merge)
        else:
            key = path[0]
            await self.access.validate_access(
                gamespace_id,
                [key],
                ProfileAccessModel.WRITE)

            result = await self.set_profile_data(gamespace_id, account_id, fields, path, merge=merge)

        return result

    async def set_profile_rw(self, gamespace_id, account_id, fields, path, merge=True):
        result = await self.set_profile_data(gamespace_id, account_id, fields, path, merge=merge)
        return result

    async def set_profiles_rw(self, gamespace_id, accounts: dict, merge=True):
        result = await self.set_profiles_data(gamespace_id, accounts, merge=merge)
        return result


class UserProfile(profile.DatabaseProfile):
    # noinspection PyShadowingNames
    @staticmethod
    def __encode_profile__(profile):
        return ujson.dumps(profile)

    def __init__(self, db, gamespace_id, account_id):
        super(UserProfile, self).__init__(db)
        self.gamespace_id = gamespace_id
        self.account_id = account_id

    # noinspection PyShadowingNames
    @staticmethod
    def __parse_profile__(profile):
        return profile

    # noinspection PyShadowingNames
    @staticmethod
    def __process_dates__(profile):
        if ProfilesModel.TIME_CREATED not in profile:
            profile[ProfilesModel.TIME_CREATED] = access.utc_time()

        profile[ProfilesModel.TIME_UPDATED] = access.utc_time()

    async def get(self):
        user = await self.conn.get(
            """
                SELECT `payload`
                FROM `account_profiles`
                WHERE `account_id`=%s AND `gamespace_id`=%s
                FOR UPDATE;
            """, self.account_id, self.gamespace_id)

        if user:
            return UserProfile.__parse_profile__(user["payload"])

        raise profile.NoDataError()

    async def insert(self, data):
        UserProfile.__process_dates__(data)
        data = UserProfile.__encode_profile__(data)

        await self.conn.insert(
            """
                INSERT INTO `account_profiles`
                (`account_id`, `gamespace_id`, `payload`)
                VALUES (%s, %s, %s);
            """, self.account_id, self.gamespace_id, data)

    async def update(self, data):
        UserProfile.__process_dates__(data)
        encoded = UserProfile.__encode_profile__(data)
        await self.conn.execute(
            """
                UPDATE `account_profiles`
                SET `payload`=%s
                WHERE `account_id`=%s AND `gamespace_id`=%s;
            """, encoded, self.account_id, self.gamespace_id)


class UserProfiles(profile.DatabaseProfile):
    # noinspection PyShadowingNames
    @staticmethod
    def __encode_profile__(profile):
        return ujson.dumps(profile)

    def __init__(self, db, gamespace_id, account_ids):
        super(UserProfiles, self).__init__(db)
        self.gamespace_id = gamespace_id
        self.account_ids = account_ids

    # noinspection PyShadowingNames
    @staticmethod
    def __process_dates__(profile):
        if ProfilesModel.TIME_CREATED not in profile:
            profile[ProfilesModel.TIME_CREATED] = access.utc_time()

        profile[ProfilesModel.TIME_UPDATED] = access.utc_time()

    async def get(self):
        users = await self.conn.query(
            """
                SELECT `payload`, `account_id`
                FROM `account_profiles`
                WHERE `account_id` IN %s AND `gamespace_id`=%s
                FOR UPDATE;
            """, self.account_ids, self.gamespace_id)

        return {
            str(user["account_id"]): user["payload"]
            for user in users
        }

    async def insert(self, data):
        # not supported since get never returns NoDataError
        pass

    async def update(self, data: dict):
        for account_id, account_profile in data.items():
            UserProfiles.__process_dates__(account_profile)

        values = []
        entries = []

        for account_id, account_profile in data.items():
            values.append("(%s, %s, %s)")
            entries.extend([account_id, self.gamespace_id,
                            UserProfiles.__encode_profile__(account_profile)])

        await self.conn.execute(
            """
                REPLACE INTO `account_profiles`
                (`account_id`, `gamespace_id`, `payload`)
                VALUES {0};
            """.format(", ".join(values)), *entries)




import anthill.common.admin as a
from anthill.common.internal import Internal, InternalError
from anthill.common.validate import validate
from anthill.common import access

from . model.access import NoAccessData
from . model.profile import ProfileError, NoSuchProfileError, ProfileQueryError

import json


class GamespaceAccessController(a.AdminController):
    async def get(self):

        access_data = self.application.access

        try:
            access = await access_data.get_access(self.gamespace)
        except NoAccessData:
            access_private = ""
            access_public = ""
            access_protected = ""
        else:
            access_private = ",".join(access.get_private())
            access_public = ",".join(access.get_public())
            access_protected = ",".join(access.get_protected())

        result = {
            "access_private": access_private,
            "access_public": access_public,
            "access_protected": access_protected
        }

        return result

    def render(self, data):
        return [
            a.breadcrumbs([], "Profile access"),
            a.form("Edit access to profile fields", fields={
                "access_public": a.field(
                    "Public access fields (everybody may see, owner may change them)",
                    "tags", "primary"),
                "access_private": a.field(
                    "Private access fields (server only fields, no one except server can see or change)",
                    "tags", "primary"),
                "access_protected": a.field(
                    "Protected access fields (only owner may see, only server may change)",
                    "tags", "primary")
            }, methods={
                "update": a.method("Update", "primary")
            }, data=data),
            a.links("Navigate", [
                a.link("@back", "Go back", icon="chevron-left")
            ])
        ]

    def access_scopes(self):
        return ["profile_admin"]

    async def update(self, access_private, access_public, access_protected):

        access_data = self.application.access

        await access_data.set_access(self.gamespace,
                                     access_private.split(","),
                                     access_protected.split(","),
                                     access_public.split(","))

        result = {
            "access_private": access_private,
            "access_public": access_public,
            "access_protected": access_protected
        }

        return result


class ProfileController(a.AdminController):
    async def get(self, account):

        profiles = self.application.profiles

        try:
            profile = (await profiles.get_profile_data(self.gamespace, account, None))
        except NoSuchProfileError:
            profile = {}

        result = {
            "profile": profile
        }

        return result

    def render(self, data):
        return [
            a.breadcrumbs([
                a.link("profiles", "Edit user profiles"),
            ], "@{0}".format(self.context.get("account"))),
            a.form(title="Edit account profile", fields={
                "profile": a.field("Profile", "json", "primary", "non-empty")
            }, methods={
                "update": a.method("Update", "primary")
            }, data=data),
            a.links("Navigate", [
                a.link("profiles", "Go back", icon="chevron-left")
            ])
        ]

    def access_scopes(self):
        return ["profile_admin"]

    async def update(self, profile):

        try:
            profile = json.loads(profile)
        except (KeyError, ValueError):
            raise a.ActionError("Corrupted profile")

        profiles = self.application.profiles
        account_id = self.context.get("account")

        try:
            await profiles.set_profile_data(self.gamespace, account_id, profile, None, merge=False)
        except ProfileError as e:
            raise a.ActionError(e.message)

        raise a.Redirect(
            "profile",
            message="Profile has been updated",
            account=account_id)


class ProfilesController(a.AdminController):
    def render(self, data):
        return [
            a.breadcrumbs([], "Edit user profile"),
            a.split([
                a.form(title="Find by credential", fields={
                    "credential": a.field("User credential", "text", "primary", "non-empty"),
                }, methods={
                    "search_credential": a.method("Search", "primary")
                }, data=data),
                a.form(title="Find by account number", fields={
                    "account": a.field("Account number", "text", "primary", "number")
                }, methods={
                    "search_account": a.method("Search", "primary")
                }, data=data)
            ]),
            a.links("Navigate", [
                a.link("index", "Go back", icon="chevron-left")
            ])
        ]

    def access_scopes(self):
        return ["profile_admin"]

    async def search_account(self, account):
        raise a.Redirect("profile", account=account)

    async def search_credential(self, credential):

        internal = Internal()

        try:
            account = await internal.request(
                "login",
                "get_account",
                credential=credential)

        except InternalError as e:
            if e.code == 400:
                raise a.ActionError("Failed to find credential: bad username")
            if e.code == 404:
                raise a.ActionError("Failed to find credential: no such user")

            raise a.ActionError(e.body)

        raise a.Redirect("profile", account=account["id"])


class QueryProfilesController(a.AdminController):
    def render(self, data):
        r = [
            a.breadcrumbs([], "Query User Profiles")
        ]

        results = data.get("results", None)
        if results is not None:
            r.extend([
                a.content(
                    "Query Results: Found {0} profile(s)".format(data.get("count", 0)),
                    [
                        {
                            "id": "account_id",
                            "title": "Account"
                        },
                        {
                            "id": "profile",
                            "title": "Profile Object"
                        }
                    ],
                    [
                        {
                            "account_id": [
                                a.link("profile", result.account, icon="user", account=result.account)
                            ],
                            "profile": [
                                a.json_view(result.profile)
                            ],
                        } for result in data["results"]
                    ],
                    "default", empty="Query yielded no results"
                )
            ])

        r.extend([
            a.form(title="Query", fields={
                "query": a.field("Profile Search Query", "json", "primary", "non-empty", height=200,
                                 description="""
                                            Please read <a href="https://docs.anthillplatform.org/en/latest/other/dbquery.html" target="_blank">this document</a>
                                            for the query format.
                                         """),
            }, methods={
                "do_query": a.method("Search", "primary")
            }, data=data),
            a.links("Navigate", [
                a.link("index", "Go back", icon="chevron-left")
            ])
        ])

        return r

    @validate(query="load_json_dict")
    async def do_query(self, query):

        if not query:
            raise a.ActionError("Query cannot be an empty object")

        profiles = self.application.profiles

        q = profiles.profile_query(self.gamespace)
        q.filters = query
        q.limit = 1000

        try:
            results, count = await q.query(count=True)
        except ProfileQueryError as e:
            raise a.ActionError(str(e))

        return {
            "results": results,
            "count": count,
            "query": query
        }

    async def get(self):
        return {
            "query": {}
        }

    def access_scopes(self):
        return ["profile_admin"]

    async def search_account(self, account):
        raise a.Redirect("profile", account=account)

    async def search_credential(self, credential):

        internal = Internal()

        try:
            account = await internal.request(
                "login",
                "get_account",
                credential=credential)

        except InternalError as e:
            if e.code == 400:
                raise a.ActionError("Failed to find credential: bad username")
            if e.code == 404:
                raise a.ActionError("Failed to find credential: no such user")

            raise a.ActionError(e.body)

        raise a.Redirect("profile", account=account["id"])


class RootAdminController(a.AdminController):
    def render(self, data):
        return [
            a.links("Profile service", [
                a.link("profiles", "Edit User Profiles", icon="user"),
                a.link("query", "Query User Profiles", icon="search"),
                a.link("access", "Edit Profile Access", icon="lock")
            ])
        ]

    def access_scopes(self):
        return ["profile_admin"]




from anthill.common.options import options
from anthill.common import server, database, access

from . model.profile import ProfilesModel
from . model.access import ProfileAccessModel
from . import handler as h
from . import options as _opts
from . import admin


class ProfileServer(server.Server):
    # noinspection PyShadowingNames
    def __init__(self):
        super(ProfileServer, self).__init__()

        self.db = database.Database(
            host=options.db_host,
            database=options.db_name,
            user=options.db_username,
            password=options.db_password)

        self.access = ProfileAccessModel(self.db)
        self.profiles = ProfilesModel(self.db, self.access)

    def get_models(self):
        return [self.access, self.profiles]

    def get_admin(self):
        return {
            "index": admin.RootAdminController,
            "access": admin.GamespaceAccessController,
            "profiles": admin.ProfilesController,
            "profile": admin.ProfileController,
            "query": admin.QueryProfilesController
        }

    def get_metadata(self):
        return {
            "title": "User Profiles",
            "description": "Manage the profiles of the users",
            "icon": "user"
        }

    def get_handlers(self):
        return [
            (r"/profile/me/?([\w/]*)", h.ProfileMeHandler),
            (r"/profile/([\w]+)/?([\w/]*)", h.ProfileUserHandler),
            (r"/profiles", h.MassProfileUsersHandler)
        ]

    def get_internal_handler(self):
        return h.InternalHandler(self)


if __name__ == "__main__":
    stt = server.init()
    access.AccessToken.init([access.public()])
    server.start(ProfileServer)




from anthill.common import handler, access
from tornado.web import HTTPError

from anthill.common.access import scoped, internal
from anthill.common.internal import InternalError
from anthill.common.validate import validate_value, ValidationError

from . model.profile import NoSuchProfileError, ProfileError, ProfileQueryError
from . model.access import AccessDenied

import ujson


class InternalHandler(object):
    def __init__(self, application):
        self.application = application

    async def mass_profiles(self, action, gamespace, accounts, profile_fields=None):

        profiles_data = self.application.profiles

        try:
            profiles = await profiles_data.get_profiles(gamespace, action, [str(account) for account in accounts],
                                                        profile_fields or [])
        except ProfileError as e:
            raise InternalError(400, "Failed to get profiles: " + e.message)
        else:
            return profiles

    async def update_profile(self, gamespace_id, account_id, fields, path="", merge=True):

        profiles = self.application.profiles

        path = filter(bool, path.split("/")) if path is not None else None

        if not isinstance(fields, dict):
            raise InternalError(400, "Expected 'data' field to be an object (a set of fields).")

        try:
            result = await profiles.set_profile_rw(
                gamespace_id,
                account_id,
                fields,
                path,
                merge=merge)

        except ProfileError as e:
            raise InternalError(400, e.message)
        except AccessDenied as e:
            raise InternalError(403, str(e))
        else:
            return result

    async def get_my_profile(self, gamespace_id, account_id, path=""):
        profiles = self.application.profiles

        path = list(filter(bool, path.split("/"))) if path is not None else None

        try:
            result = await profiles.get_profile_me(
                gamespace_id,
                account_id,
                path)

        except ProfileError as e:
            raise InternalError(400, e.message)
        except NoSuchProfileError:
            raise InternalError(404, "No profile found")
        except AccessDenied as e:
            raise InternalError(403, str(e))
        else:
            return result

    async def get_profile_others(self, gamespace_id, account_id, path=""):
        profiles = self.application.profiles

        path = list(filter(bool, path.split("/"))) if path is not None else None

        try:
            result = await profiles.get_profile_others(
                gamespace_id,
                account_id,
                path)

        except ProfileError as e:
            raise InternalError(400, e.message)
        except NoSuchProfileError:
            raise InternalError(404, "No profile found")
        except AccessDenied as e:
            raise InternalError(403, str(e))
        else:
            return result

    async def query_profiles(self, gamespace_id, query, limit=1000):
        profiles = self.application.profiles

        q = profiles.profile_query(gamespace_id)
        q.filters = query
        q.limit = limit

        try:
            results, count = await q.query(count=True)
        except ProfileQueryError as e:
            raise InternalError(500, str(e))

        return {
            "results": {
                r.account: {
                    "profile": r.profile
                }
                for r in results
            },
            "total_count": count
        }


class ProfileMeHandler(handler.AuthenticatedHandler):
    @scoped(scopes=["profile"])
    async def get(self, path):

        profiles = self.application.profiles

        account_id = self.current_user.token.account

        gamespace_id = self.current_user.token.get(
            access.AccessToken.GAMESPACE)

        path = list(filter(bool, path.split("/"))) if path is not None else None

        if self.token.has_scope("profile_private"):
            method = profiles.get_profile_data
        else:
            method = profiles.get_profile_me

        try:
            profile = await method(gamespace_id, account_id, path)

        except NoSuchProfileError:
            raise HTTPError(404, "Profile was not found.")
        except AccessDenied:
            raise HTTPError(403, "Access denied")
        else:
            self.dumps(profile)

    @scoped(scopes=["profile_write"])
    async def post(self, path):

        profiles = self.application.profiles

        account_id = self.current_user.token.account

        gamespace_id = self.current_user.token.get(
            access.AccessToken.GAMESPACE)

        path = list(filter(bool, path.split("/"))) if path is not None else None

        try:
            fields = ujson.loads(self.get_argument("data"))
        except (KeyError, ValueError):
            raise HTTPError(400, "Corrupted 'data' field: expecting JSON object.")

        if not isinstance(fields, dict):
            raise HTTPError(400, "Expected 'data' field to be an object (a set of fields).")

        merge = self.get_argument("merge", "true") == "true"

        if self.token.has_scope("profile_private"):
            method = profiles.set_profile_rw
        else:
            method = profiles.set_profile_me

        try:
            result = await method(
                gamespace_id,
                account_id,
                fields,
                path,
                merge=merge)

        except ProfileError as e:
            raise HTTPError(400, str(e))
        except AccessDenied as e:
            raise HTTPError(403, str(e))
        else:
            self.dumps(result)


class ProfileUserHandler(handler.AuthenticatedHandler):
    @scoped(scopes=["profile"])
    async def get(self, account_id, path):

        profiles = self.application.profiles

        gamespace_id = self.current_user.token.get(access.AccessToken.GAMESPACE)

        path = list(filter(bool, path.split("/"))) if path is not None else None

        try:
            profile = await profiles.get_profile_others(gamespace_id, account_id, path)

        except NoSuchProfileError:
            raise HTTPError(404, "Profile was not found.")
        else:
            self.dumps(profile)

    @internal
    async def post(self, account_id, path):

        profiles = self.application.profiles

        gamespace_id = self.current_user.token.get(access.AccessToken.GAMESPACE)

        path = list(filter(bool, path.split("/"))) if path is not None else None

        try:
            fields = ujson.loads(self.get_argument("data"))
        except (KeyError, ValueError):
            raise HTTPError(400, "Corrupted 'data' field: expecting JSON object.")

        if not isinstance(fields, dict):
            raise HTTPError(400, "Expected 'data' field to be an object (a set of fields).")

        merge = self.get_argument("merge", True)

        try:
            result = await profiles.set_profile_rw(gamespace_id, account_id, fields, path, merge=merge)

        except ProfileError as e:
            raise HTTPError(400, str(e))
        except AccessDenied as e:
            raise HTTPError(403, str(e))
        else:
            self.dumps(result)


class MassProfileUsersHandler(handler.AuthenticatedHandler):
    @scoped(scopes=["profile"])
    async def get(self):

        try:
            accounts = ujson.loads(self.get_argument("accounts"))
        except (KeyError, ValueError):
            raise HTTPError(400, "Corrupted 'accounts' field.")

        try:
            accounts = validate_value(accounts, "json_list_of_ints")
        except ValidationError as e:
            raise HTTPError(400, e.message)

        profile_fields = self.get_argument("profile_fields", None)

        if profile_fields:

            try:
                profile_fields = ujson.loads(profile_fields)
                profile_fields = validate_value(profile_fields, "json_list_of_strings")
            except (KeyError, ValueError, ValidationError):
                raise HTTPError(400, "Corrupted profile_fields")

        if len(accounts) > 100:
            raise HTTPError(400, "To many accounts to request.")

        profiles_data = self.application.profiles
        gamespace_id = self.current_user.token.get(access.AccessToken.GAMESPACE)

        try:
            profiles = await profiles_data.get_profiles(
                gamespace_id, "get_public", [str(account) for account in accounts],
                profile_fields or [])
        except ProfileError as e:
            raise HTTPError(400, "Failed to get profiles: " + e.message)
        else:
            self.dumps(profiles)

    @scoped(scopes=["profile", "profile_write", "profile_multi"])
    async def post(self):
        profiles_data = self.application.profiles
        gamespace_id = self.current_user.token.get(access.AccessToken.GAMESPACE)

        try:
            profiles = ujson.loads(self.get_argument("data"))
            profiles = validate_value(profiles, "json_dict_of_dicts")
        except (KeyError, ValueError, ValidationError):
            raise HTTPError(400, "Corrupted 'data' field: expecting JSON object of JSON objects.")

        merge = self.get_argument("merge", True)

        try:
            result = await profiles_data.set_profiles_rw(gamespace_id, profiles, merge=merge)

        except ProfileError as e:
            raise HTTPError(400, str(e))
        except AccessDenied as e:
            raise HTTPError(403, str(e))
        else:
            self.dumps(result)



from anthill.common.model import Model

__author__ = "desertkun"


class AccessAdapter(object):
    def __init__(self, data):
        self.private = data.get("access_private", "").split("\n")
        self.protected = data.get("access_protected", "").split("\n")
        self.public = data.get("access_public", "").split("\n")

    def get_private(self):
        return self.private

    def get_protected(self):
        return self.protected

    def get_public(self):
        return self.public


class AccessDenied(Exception):
    pass


class NoAccessData(Exception):
    pass


class ProfileAccessModel(Model):
    READ = 0
    READ_OTHERS = 1
    WRITE = 2

    async def __get_access_data__(self, gamespace_id):

        access = await self.db.get(
            """
                SELECT * 
                FROM `gamespace_access`
                WHERE `gamespace_id`=%s;
            """, gamespace_id, cache_hash=('profile_access', gamespace_id), cache_time=600)

        if access is None:
            raise NoAccessData()

        return AccessAdapter(access)

    def __init__(self, db):
        self.db = db

    async def setup_table_gamespace_access(self):
        await self.set_access(1, [], [], ["name", "avatar", "@time_updated", "@time_created"])

    def get_setup_tables(self):
        return ["gamespace_access"]

    def get_setup_db(self):
        return self.db

    async def get_access(self, gamespace_id):
        
        try:
            access = await self.__get_access_data__(gamespace_id)
        except NoAccessData:
            return AccessAdapter({})

        return access

    async def set_access(self, gamespace_id, access_private, access_protected, access_public):

        data_private = "\n".join(access_private)
        data_protected = "\n".join(access_protected)
        data_public = "\n".join(access_public)

        try:
            await self.__get_access_data__(gamespace_id)
        except NoAccessData:
            await self.db.insert(
                """
                    INSERT INTO `gamespace_access`
                    (gamespace_id, access_private, access_protected, access_public)
                    VALUES (%s, %s, %s, %s);
                """, gamespace_id, data_private, data_protected,
                data_public, cache_hash=('profile_access', gamespace_id))

        else:
            await self.db.execute(
                """
                    UPDATE `gamespace_access`
                    SET `access_private`=%s, `access_protected`=%s, `access_public`=%s
                    WHERE `gamespace_id`=%s;
                """, data_private, data_protected, data_public,
                gamespace_id, cache_hash=('profile_access', gamespace_id))

    async def validate_access(self, gamespace_id, fields, operation):

        access = await self.get_access(gamespace_id)

        if operation == ProfileAccessModel.READ:

            private = access.get_private()
            result = list(set(fields) - set(private))
            return result

        elif operation == ProfileAccessModel.READ_OTHERS:

            public = access.get_public()
            result = list(set(public) & set(fields))
            return result

        elif operation == ProfileAccessModel.WRITE:

            private = access.get_private()
            protected = access.get_protected()

            if any((field in protected) or (field in private) for field in fields):
                raise AccessDenied()
